package Data::Collection;

#Author: Obiora Embry
#Date started: 11 August 2008

use strict;
no strict "refs";
use warnings;
use Fcntl ':flock'; #import LOCK_* constants
use File::Flock qw( lock );
use File::Copy;
use Cwd;
use Archive::Extract;
use Archive::Zip qw( :ERROR_CODES :CONSTANTS );

require Exporter;
use vars qw(@EXPORT_OK @ISA $file_hash $dir_hash);

@ISA = qw(Exporter);
@EXPORT_OK = qw(read_directory);

our $VERSION = '0.2.9';

$Archive::Extract::PREFER_BIN = 1;
$Archive::Extract::WARN = 0;

sub new 
{

my $class = shift;
		
my $self = (bless {}, $class);

$self->{OPTS} = $_[2];

$self->{help} = $self->{OPTS}->{help}; 
$self->show_help if ( $self->{help} );

$self->{version} = $self->{OPTS}->{version}; 
$self->show_version($class) if ( $self->{version} );

$self->{searchdir} = $_[0] || die qq{location not specified\n};

$self->{disk_name} = $_[1] || die qq{no name specified\n};

return $self;

}

sub show_help
{

my ($self) = shift;

print qq{Usage (see below):\n\n};
print qq{user can index and record all files on a computer disk (floppy, optical, zip, USB, etc.)\n\n};
print qq{\texample \#1:\n};
print qq{perl index_disk_cl.pl [-l disk location] [-n disk name] [-e verbose] [-d decompress]\n\n}; 
print qq{perl index_disk_cl.pl [--disk_location disk location] [--disk_name disk name] [--verbose verbose] [--decompress decompress]\n\n}; 
		
print qq{\t--help/-h: prints usage\n\n};
print qq{\t--version/-v: prints version\n\n};
print qq{\t--disk_location <disk location>/-l <disk location>: the full path of the computer disk\n\n};
print qq{\t--disk_name <disk name>/-n <disk name>: the name of the disk\n\n};
print qq{\t--verbose <verbose>/-e <verbose>: specify if you want to see all of the documents and files as they are being found [true or false]\n\n};
print qq{\t--decompress <decompress>/-d <decompress>: specify if you want to decompress compressed files [true or false]\n\n};

exit(0);

}

sub show_version
{

my ($self, $class) = @_;

print qq{\n$class version $VERSION\n};

exit(0);

}	

sub read_directory
{

my ($self) = shift;

my $disk_name = $self->{disk_name};
my $searchdir = $self->{searchdir};

my ($sem_file, $index_file) = $self->_create_index_file('');
$self->_first_lines;

my ($files, $dir) = $self->_find_files($searchdir, 0);

unless ( ref($files) =~ /HASH/ ) 
{
	
	system("mount $searchdir");
	($files, $dir) = $self->_find_files($searchdir, 1);

}

die qq{\nNo removable disk was found, please insert one!\n} unless ( ref($files) =~ /HASH/ );

$self->_last_line;
$self->_close_index_file($sem_file);

$self->_remove_duplicates_index($index_file);

my ($statement, $file_name) = $self->_create_db(\$disk_name, $files, $dir);

return ($statement);

}

sub _find_files
{

my ($self, $searchdir, $dircount) = @_;

my ($filename, $file_type) = '';
my $dirhandle = 'DIRHANDLE'.$dircount;

print qq{Reading directory $searchdir \n\n} if ( $self->{OPTS}->{verbose} && $self->{OPTS}->{verbose} == 1 );

opendir ($dirhandle, $searchdir) or die qq{Could not open directory $searchdir\n};

	while (defined($filename = readdir($dirhandle)))
	{

			if ( $filename !~ /^[.]+/ )
             	{	

				$self->_extract_archive_zip(\qq{$searchdir/$filename}, \$filename) if ( $filename =~ /(\.zip)$/ && $self->{OPTS}->{decompress} && $self->{OPTS}->{decompress} == 1 );

				$self->_extract_archive_other(\qq{$searchdir/$filename}, \$filename) if ( $filename =~ /(\.tar|\.tar\.gz|\.tgz|\.gz|\.bz2|\.tar\.bz2|\.tbz|\.lzma)$/ && $self->{OPTS}->{decompress} && $self->{OPTS}->{decompress} == 1 );

				$self->_add_file(\$searchdir, \$filename);				

				$self->_find_files(qq{$searchdir/$filename}, $dircount + 1) if (-d qq{$searchdir/$filename} );
			}
	}
		
closedir($dirhandle);

return ($file_hash, $dir_hash);

}

sub _add_file
{

my ($self, $searchdir, $filename, $archive) = @_;

my $archived = ( ref($archive) =~ /SCALAR/ ? $$archive : undef );

print qq{$$searchdir/$$filename\n\n} if ( $self->{OPTS}->{verbose} && $self->{OPTS}->{verbose} == 1 );

my ($filename_mod, $search_mod, $mode, $size, $last_mod) = ('', '', '', '', 0, 0);

$mode = (stat(qq{$$searchdir/$$filename}))[2];
$size = (stat(qq{$$searchdir/$$filename}))[7];
$last_mod = (stat(qq{$$searchdir/$$filename}))[9];

$search_mod = $$searchdir;
$search_mod =~ s!$self->{searchdir}!!;

$filename_mod = $$filename;
$filename_mod =~ s!\.$$filename!$$filename!;

my $file = qq{$search_mod/$filename_mod};
$file =~ s!\/\/!\/!;
$file =~ s!\/\.\/!\/!;

	if ( !-d qq{$$searchdir/$$filename} ) 
	{

		$file_hash->{$filename_mod}->{'file_name'} = $file;
		$file_hash->{$filename_mod}->{'archive'} = $archived if ( $archived );	
		$file_hash->{$filename_mod}->{'mode'} = sprintf("%04o", $mode & 07777) if ( $mode );
		$file_hash->{$filename_mod}->{'size'} = sprintf("%0.2f", $size / 1024) if ( $size );
 		$file_hash->{$filename_mod}->{'last_mod'} = localtime($last_mod) if ( $last_mod );

	}

	else
	{	

		$dir_hash->{$filename_mod}->{'dir_name'} = qq{$search_mod/$filename_mod};
		$dir_hash->{$filename_mod}->{'mode'} = sprintf("%04o", $mode & 07777) if ( $mode );
		$dir_hash->{$filename_mod}->{'size'} = sprintf("%0.2f", $size / 1024) if ( $size );
 		$dir_hash->{$filename_mod}->{'last_mod'} = localtime($last_mod) if ( $last_mod );

	}										

$self->_create_index($filename);

return;

}


sub _create_db
{

my ($self, $disk_name_ref, $found_files, $found_dir) = @_;

my $disk_name = $$disk_name_ref;

my $file = $self->get_xml_file($disk_name_ref);

my $cd = getcwd;

$self->_mk_dir(\qq{$cd/removable_disk});
$self->_mk_dir(\qq{$cd/removable_disk/xml});

my @alphanumeric = ('a'..'z', 'A'..'Z', 0..9);
my $rand_alphanumeric = join ('', map $alphanumeric[rand @alphanumeric], 0..9);

my $SEMAPHORE = $disk_name.$rand_alphanumeric.'.lck';
open(S, ">", "$SEMAPHORE") or die qq{Could not create $SEMAPHORE: $!\n};
flock(S, LOCK_EX) or die qq{flock() failed for $SEMAPHORE: $!\n};			
open(XML, ">", "$$file") or die qq{Could not create file name $$file\n};

$self->_xml_header;

my ($num, $num_dir) = 0;

foreach ( sort keys %$found_files )
{

	print XML qq{<file>\n};
	print XML qq{<file_name>$found_files->{$_}->{'file_name'}</file_name>\n} if ( $found_files->{$_}->{'file_name'} );
	print XML qq{<archive_name>$found_files->{$_}->{'archive'}</archive_name>\n} if ( $found_files->{$_}->{'archive'} );	
	print XML qq{<mode>$found_files->{$_}->{'mode'}</mode>\n} if ( $found_files->{$_}->{'mode'} );
	print XML qq{<size>$found_files->{$_}->{'size'} KB</size>\n} if ( $found_files->{$_}->{'size'} );
	print XML qq{<last_modified>$found_files->{$_}->{'last_mod'}</last_modified>\n} if ( $found_files->{$_}->{'last_mod'} );
	print XML qq{</file>\n};

	$num++;

}

foreach ( keys %$found_dir ) { $num_dir++; }

$self->_xml_footer;
close(XML);
close(S);

$self->_delete_file(\$SEMAPHORE);

return (\sprintf(qq{\n\t$num file%s and $num_dir director%s were found.  The database can be accessed manually at $$file.\n}, ( $num != 1 ? 's' : ''), ( $num_dir != 1 ? 'ies' : 'y' )), $file);

}

sub _xml_header
{

my ($self) = shift;

my $disk_name = $self->_get_disk_name;

print XML qq{<?xml version="1.0" encoding="UTF-8"?>\n\n};
print XML qq{<removable_disk>\n<rd_$$disk_name>\n};

}

sub _xml_footer
{

my ($self) = shift;

my $disk_name = $self->_get_disk_name;

print XML qq{</rd_$$disk_name>\n</removable_disk>};

}

sub get_xml_file
{

my ($self, $disk) = @_;

my $disk_name = $self->_get_disk_name($disk);

my $cd = getcwd;
my $xml_file = qq{$cd/removable_disk/xml/$$disk_name.xml};

return \$xml_file;

}

sub _get_disk_name
{

my ($self, $disk) = @_;

my $disk_name = ( $disk && ref($disk) =~ /SCALAR/ ? $$disk : $self->{disk_name} ); 
$disk_name =~ s/ /_/g;
$disk_name =~ s/(\(|\))//g;

$self->{disk_name} = $disk_name;

return \$disk_name;

}

sub _ignore_words
{

my $text = 'after can her may on them way all come hers me only then we also did hid more onto there were am do him most or these what an a does his much other they when and each how must out this where any etc however my over those whether are far ie near per til which as few if new put to who at fix in next same too why be for into no say try will because from is none since under with been get it nor so unto within before go its not some up without between got just now such upon yet big had led of than us both has less off that very but have let oh the vs by he many old their was';

my $ignore;

my @words2 = split (/\s+/, $text);
map { $ignore .= '\b'.$_.'\b|' } @words2;

$ignore =~ s/\|$//g if ( $ignore );

return \$ignore;

}

sub _create_index
{

my ($self, $filename_ref) = @_;

my $ignore = $self->_ignore_words;

my ($i, @words2, @words) = ( 0, (), () );

my $filename = $$filename_ref;
$filename =~ s!_! !g;
$filename =~ s/\.[a-zA-Z]{1,4}$//;

@words = split(/\s+/, $filename);

$filename =~ s/ /_/g;

for ( $i = 0; $i < @words; $i++ )
{

	$words[$i] =~ s/ |%/_/g;
	$words[$i] =~ s/&//g;
	$words[$i] =~ s/-//g;
	$words[$i] =~ s/^\d*$//g;
	$words[$i] =~ s/\(|\)//g;

	push (@words2, $words[$i]) if ( $words[$i] !~ /$$ignore/ig && $words[$i] !~ // );

}

map { print INDEX qq{<index>$_</index>\n} } @words2;

}

sub _create_index_file
{

my ($self, $num) = @_;

$self->_mk_dir(\'rem_disk_index');

my $disk_name = $self->_get_disk_name;
my $file = 'rem_disk_index/'.$$disk_name.qq{_index$num.xml};

my @alphanumeric = ('a'..'z', 'A'..'Z', 0..9);
my $rand_alphanumeric = join ('', map $alphanumeric[rand @alphanumeric], 0..9);

my $SEMAPHORE = $file.$rand_alphanumeric.'.lck';
open(SINDEX, ">", "$SEMAPHORE") or die qq{Could not create $SEMAPHORE: $!\n};
flock(SINDEX, LOCK_EX) or die qq{flock() failed for $SEMAPHORE: $!\n};		
open(INDEX, ">", "$file") or die qq{Could not create file name $file\n};

return (\$SEMAPHORE, \$file);

}

sub _first_lines
{

my ($self) = shift;

print INDEX qq{<?xml version="1.0" encoding="UTF-8"?>\n\n};
print INDEX qq{<removable_disk>\n};

}

sub _last_line
{

my ($self) = shift;

print INDEX qq{</removable_disk>};

}

sub _close_index_file
{

my ($self, $sem_file) = @_;

close(INDEX);
close(SINDEX);

$self->_delete_file($sem_file);

}

sub _remove_duplicates_index
{

my ($self, $file) = @_;

my %seen = ();

open(INDEX4, "<", "$$file") or die qq{Couldn't open $$file for reading!\n};
lock($$file, 'shared');
my ($sem_file, $file2) = $self->_create_index_file(2); 

while ( <INDEX4> )
{

	$seen{$_}++;
	next if ( $seen{$_} > 1 );
	print INDEX $_;
      
}

close(INDEX4);
$self->_close_index_file($sem_file);

move($$file2, $$file);

$self->_delete_file($sem_file);
$self->_delete_file($file2);

}


sub _mk_dir
{

my ($self, $dir) = @_;

die qq{dir not given\n} unless ( ref($dir) =~ /SCALAR/ );

unless ( -d $$dir )
{

	eval { mkdir $$dir };

	die qq{could not make directory $$dir, $@\n} if ( $@ );

}

}

sub _delete_file
{

my ($self, $file) = @_;

eval { unlink $$file };

warn qq{Could not delete file: $$file: $!\n} if ( $@ );

}

sub _extract_archive_other
{

my ($self, $archive_ref, $archive) = @_;

my ($temp_archive_dir, $temp_dir) = ('temp_archive', 'temp');

$self->_mk_dir(\$temp_archive_dir);
$self->_mk_dir(\$temp_dir);

copy($$archive_ref, $temp_archive_dir) or warn qq{\nCopy failed: $!\n};

my $filename = qq{$temp_archive_dir/$$archive};
my $filename_wo = $filename;

$filename_wo =~ s/ /_/g;

move($filename, $filename_wo) unless ( $filename eq $filename_wo );

my $ae = Archive::Extract->new( archive => $filename_wo );

my $error = $ae->extract( to => $temp_dir );

if ( $error )
{

	my $cwd = cwd;
	my $path = qq{$cwd/$temp_dir};
	
	foreach ( @{$ae->files} )
	{
	
		$self->_add_file($archive_ref, \$_, $archive);

		$self->_delete_file(\qq{$path/$_}) if ( -e qq{$path/$_} && -f qq{$path/$_} );
		rmdir qq{$path/$_} if ( -d qq{$path/$_} );

	}

}

$self->_delete_file(\$filename_wo) if ( -e $filename_wo );

return;

}

sub _extract_archive_zip
{

my ($self, $archive_ref, $archive) = @_;

my ($zip, $element) = ( '', '' );

$zip = Archive::Zip->new;

return unless ( $zip->read( $$archive_ref ) == AZ_OK );

return if $zip->diskNumberWithStartOfCentralDirectory; #check for Multi-volume zip file

my ($temp_archive_dir, $temp_dir) = ('temp_archive', 'temp');

$self->_mk_dir(\$temp_archive_dir);
$self->_mk_dir(\$temp_dir);

copy($$archive_ref, $temp_archive_dir) or warn qq{\nCopy failed: $!\n};

my $filename = qq{$temp_archive_dir/$$archive};
my $filename_wo = $filename;

$filename_wo =~ s/ /_/g;

move($filename, $filename_wo) unless ( $filename eq $filename_wo );

my $cwd = cwd;
my $path = qq{$cwd/$temp_dir};
	
	foreach $element ( @{$zip->members()} ) 
	{

		next if $element->isEncrypted;

		$self->_add_file($archive_ref, \$element, $archive);

		$self->_delete_file(\qq{$path/$element}) if ( -e qq{$path/$element} && -f qq{$path/$element} );
		rmdir qq{$path/$element} if ( -d qq{$path/$element} );

	}

$self->_delete_file(\$filename_wo) if ( -e $filename_wo );

return;

}

sub DESTROY
{

my ($self) = shift;

system("umount $self->{searchdir} ") if ( $self->{searchdir} );

}

1;

__END__
